@@ -57,8 +57,6 @@ struct _GnomePrintJobPreview {
 	/* Some interesting buttons */
 	GtkWidget *bpf, *bpp, *bpn, *bpl;
 	GtkWidget *bz1, *bzf, *bzi, *bzo;
-	/* Number of pages job has */
-	gint pages;
 	/* Zoom factor */
 	gdouble zoom;
 
@@ -73,6 +71,12 @@ struct _GnomePrintJobPreview {
 	gint offsetx, offsety;
 
 	gpointer priv;
+
+	/* Number of pages displayed together */
+	gulong nx, ny;
+	gint update_func_id;
+
+	GPtrArray *page_array;
 };
 
 struct _GnomePrintJobPreviewClass {
@@ -101,6 +105,7 @@ struct _GPMPPrivate {
 };
 
 static void gpmp_parse_layout (GnomePrintJobPreview *mp);
+static void gnome_print_job_preview_update (GnomePrintJobPreview *pmp);
 
 /*
  * Padding in points around the simulated page
@@ -108,65 +113,36 @@ static void gpmp_parse_layout (GnomePrintJobPreview *mp);
 
 #define PAGE_PAD 4
 
-static gint
-render_page (GnomePrintJobPreview *mp, gint page)
-{
-	GPMPPrivate *priv;
-	GnomePrintConfig *config;
-	ArtDRect region;
-	gdouble transform[6];
-
-	priv = (GPMPPrivate *) mp->priv;
-
-	if (priv->preview) {
-		g_object_unref (G_OBJECT (priv->preview));
-		priv->preview = NULL;
-	}
-
-	/* Set page transformation */
-	transform[0] =  1.0;
-	transform[1] =  0.0;
-	transform[2] =  0.0;
-	transform[3] = -1.0;
-	transform[4] =  0.0;
-	transform[5] =  mp->pah;
-	art_affine_multiply (transform, mp->pa2ly, transform);
-
-	/* Reset scrolling region always */
-	region.x0 = region.y0 = 0.0 - PAGE_PAD;
-	region.x1 = mp->paw + PAGE_PAD;
-	region.y1 = mp->pah + PAGE_PAD;
-
-	config = gnome_print_job_get_config (priv->job);
-	priv->preview = gnome_print_preview_new_full (config, priv->canvas, transform, &region);
-	gnome_print_preview_theme_compliance (GNOME_PRINT_PREVIEW (priv->preview),
-					      priv->theme_compliance);
-	gnome_print_config_unref (config);
-
-	gnome_print_job_render_page (priv->job, priv->preview, page, TRUE);
-
-	return GNOME_PRINT_OK;
-}
-
 static gint
 goto_page (GnomePrintJobPreview *mp, gint page)
 {
 	GPMPPrivate *priv;
 	guchar c[32];
 
+	g_return_val_if_fail (mp != NULL, GNOME_PRINT_ERROR_BADVALUE);
+
 	priv = (GPMPPrivate *) mp->priv;
 
+	g_return_val_if_fail ((priv->pagecount == 0) || 
+			      (page < priv->pagecount),
+			      GNOME_PRINT_ERROR_BADVALUE);
+
 	g_snprintf (c, 32, "%d", page + 1);
 	gtk_entry_set_text (GTK_ENTRY (priv->page_entry), c);
 
-	gtk_widget_set_sensitive (mp->bpf, (page != 0) && priv->pagecount > 1);
-	gtk_widget_set_sensitive (mp->bpp, (page != 0) && priv->pagecount > 1);
-	gtk_widget_set_sensitive (mp->bpn, (page != (mp->pages - 1)) && priv->pagecount > 1);
-	gtk_widget_set_sensitive (mp->bpl, (page != (mp->pages - 1)) && priv->pagecount > 1);
+	gtk_widget_set_sensitive (mp->bpf, (page > 0) &&
+					   (priv->pagecount > 1));
+	gtk_widget_set_sensitive (mp->bpp, (page > 0) &&
+					   (priv->pagecount > 1));
+	gtk_widget_set_sensitive (mp->bpn, (page != (priv->pagecount - 1)) &&
+					   (priv->pagecount > 1));
+	gtk_widget_set_sensitive (mp->bpl, (page != (priv->pagecount - 1)) &&
+					   (priv->pagecount > 1));
 
 	if (page != priv->current_page) {
 		priv->current_page = page;
-		return render_page (mp, page);
+		if (priv->pagecount > 0)
+			gnome_print_job_preview_update (mp);
 	}
 
 	return GNOME_PRINT_OK;
@@ -201,7 +177,7 @@ gpmp_zoom (GnomePrintJobPreview *mp, gdouble factor, gboolean relative)
 	if (relative) {
 		zoom = mp->zoom * factor;
 	} else {
-		zoom = factor;
+		zoom = factor / MAX (mp->nx, mp->ny);
 	}
 
 	mp->zoom = CLAMP (zoom, GPMP_ZOOM_MIN, GPMP_ZOOM_MAX);
@@ -224,6 +200,7 @@ preview_canvas_button_press (GtkWidget *widget, GdkEventButton *event, GnomePrin
 
 	if (event->button == 1) {
 		GdkCursor *cursor;
+		GdkDisplay *display = gtk_widget_get_display (widget);
 
 		mp->dragging = TRUE;
 
@@ -232,7 +209,7 @@ preview_canvas_button_press (GtkWidget *widget, GdkEventButton *event, GnomePrin
 		mp->anchorx = event->x - mp->offsetx;
 		mp->anchory = event->y - mp->offsety;
 
-		cursor = gdk_cursor_new (GDK_FLEUR);
+		cursor = gdk_cursor_new_for_display (display, GDK_FLEUR);
 		gdk_pointer_grab (widget->window, FALSE,
 				  (GDK_POINTER_MOTION_MASK | 
 				   GDK_POINTER_MOTION_HINT_MASK | 
@@ -293,7 +270,8 @@ preview_canvas_button_release (GtkWidget *widget, GdkEventButton *event, GnomePr
 
 	if (event->button == 1) {
 		mp->dragging = FALSE;
-		gdk_pointer_ungrab (event->time);
+		gdk_display_pointer_ungrab (gtk_widget_get_display (widget),
+					    event->time);
 		retval = TRUE;
 	}
 
@@ -336,33 +314,25 @@ preview_first_page_cmd (void *unused, GnomePrintJobPreview *pmp)
 static void
 preview_next_page_cmd (void *unused, GnomePrintJobPreview *pmp)
 {
-	GPMPPrivate *priv;
+	GPMPPrivate *priv = (GPMPPrivate *) pmp->priv;
 
-	priv = (GPMPPrivate *) pmp->priv;
-
-	if (priv->current_page < (priv->pagecount - 1)) {
-		goto_page (pmp, priv->current_page + 1);
-	}
+	goto_page (pmp, MIN (priv->current_page + pmp->nx * pmp->ny,
+			     priv->pagecount - 1));
 }
 
 static void
 preview_prev_page_cmd (void *unused, GnomePrintJobPreview *pmp)
 {
-	GPMPPrivate *priv;
+	GPMPPrivate *priv = (GPMPPrivate *) pmp->priv;
 
-	priv = (GPMPPrivate *) pmp->priv;
-
-	if (priv->current_page > 0) {
-		goto_page (pmp, priv->current_page - 1);
-	}
+	goto_page (pmp,
+		   MAX ((gint) (priv->current_page - pmp->nx * pmp->ny), 0));
 }
 
 static void
 preview_last_page_cmd (void *unused, GnomePrintJobPreview *pmp)
 {
-	GPMPPrivate *priv;
-
-	priv = (GPMPPrivate *) pmp->priv;
+	GPMPPrivate *priv = (GPMPPrivate *) pmp->priv;
 
 	goto_page (pmp, priv->pagecount - 1);
 }
@@ -391,8 +361,8 @@ preview_zoom_fit_cmd (GtkToggleButton *t, GnomePrintJobPreview *mp)
 	width = GTK_WIDGET (priv->canvas)->allocation.width;
 	height = GTK_WIDGET (priv->canvas)->allocation.height;
 
-	zoomx = width / (mp->paw + 5.0 + PAGE_PAD);
-	zoomy = height / (mp->pah + 5.0 + PAGE_PAD);
+	zoomx = width / (mp->paw + 5.0 + PAGE_PAD * 2);
+	zoomy = height / (mp->pah + 5.0 + PAGE_PAD * 2);
 
 	gpmp_zoom (mp, MIN (zoomx, zoomy), FALSE);
 }
@@ -509,6 +479,13 @@ preview_canvas_key (GtkWidget *widget, GdkEventKey *event, gpointer data)
 	case GDK_Escape:
 		gtk_widget_destroy (GTK_WIDGET (pmp));
 		return TRUE;
+	case GDK_q:
+	case GDK_w:
+		if (event->state & GDK_CONTROL_MASK) {
+			gtk_widget_destroy (GTK_WIDGET (pmp));
+			return TRUE;		
+		}
+	        /* break skipped */
 	default:
 		return FALSE;
 	}
@@ -528,6 +505,8 @@ canvas_style_changed_cb (GtkWidget *canvas, GtkStyle *ps, GnomePrintJobPreview *
 	GtkStyle *style;
 	gint32 border_color;
 	gint32 page_color;
+	guint i;
+	GnomeCanvasItem *page;
 
 	priv = (GPMPPrivate *) mp->priv;
 
@@ -535,10 +514,13 @@ canvas_style_changed_cb (GtkWidget *canvas, GtkStyle *ps, GnomePrintJobPreview *
 	page_color = GPP_COLOR_RGBA (style->base [GTK_STATE_NORMAL], 0xff);
 	border_color = GPP_COLOR_RGBA (style->text [GTK_STATE_NORMAL], 0xff);
 
-	gnome_canvas_item_set (priv->page,
+	for (i = 0; i < mp->page_array->len; i++) {
+		page = g_object_get_data (mp->page_array->pdata[i], "page");
+		gnome_canvas_item_set (page,
 			       "fill_color_rgba", page_color,
 			       "outline_color_rgba", border_color,
 			       NULL);
+	}
 }
 
 static void
@@ -593,8 +575,6 @@ entry_focus_out_event_cb (GtkWidget *widget, GdkEventFocus *event, GnomePrintJob
 static void
 create_preview_canvas (GnomePrintJobPreview *mp)
 {
-	GnomePrintConfig *config;
-	GnomeCanvasItem *item;
 	AtkRelationSet *relation_set;
 	AtkRelation *relation;
 	GPMPPrivate *priv;
@@ -606,6 +586,9 @@ create_preview_canvas (GnomePrintJobPreview *mp)
 	priv = (GPMPPrivate *) mp->priv;
 
 	priv->scrolled_window = gtk_scrolled_window_new (NULL, NULL);
+	gtk_scrolled_window_set_policy (
+		GTK_SCROLLED_WINDOW (priv->scrolled_window),
+		GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
 
 	gtk_widget_push_colormap (gdk_rgb_get_colormap ());
 	priv->canvas = GNOME_CANVAS (gnome_canvas_new_aa ());
@@ -625,43 +608,11 @@ create_preview_canvas (GnomePrintJobPreview *mp)
 
 	gtk_container_add (GTK_CONTAINER (priv->scrolled_window), GTK_WIDGET (priv->canvas));
 
-	config = gnome_print_job_get_config (priv->job);
-	priv->preview = gnome_print_preview_new (config, priv->canvas);
-	gnome_print_config_unref (config);
-
 	if (priv->theme_compliance) {
 		g_signal_connect (G_OBJECT (priv->canvas), "style_set",
 				  G_CALLBACK (canvas_style_changed_cb), mp);
 	}
 	
-	/*
-	 * Now add some padding above and below and put a simulated
-	 * page on the background
-	 */
-	item = gnome_canvas_item_new (gnome_canvas_root (priv->canvas),
-				      GNOME_TYPE_CANVAS_RECT,
-				      "x1", 0.0,
-				      "y1", 0.0,
-				      "x2", (gdouble) mp->paw,
-				      "y2", (gdouble) mp->pah,
-				      "fill_color", "white",
-				      "outline_color", "black",
-				      "width_pixels", 1,
-				      NULL);
-	priv->page = item;
-	gnome_canvas_item_lower_to_bottom (item);
-	item = gnome_canvas_item_new (gnome_canvas_root (priv->canvas),
-				      GNOME_TYPE_CANVAS_RECT,
-				      "x1", 3.0,
-				      "y1", 3.0,
-				      "x2", (gdouble) mp->paw + 3,
-				      "y2", (gdouble) mp->pah + 3,
-				      "fill_color", "black",
-				      NULL);
-	gnome_canvas_item_lower_to_bottom (item);
-	gnome_canvas_set_scroll_region (priv->canvas, 0 - PAGE_PAD, 0 - PAGE_PAD, 
-					mp->paw + PAGE_PAD, mp->pah + PAGE_PAD);
-
 	status = gtk_hbox_new (FALSE, 0);
 	l = gtk_label_new_with_mnemonic (_("P_age: "));
 	gtk_box_pack_start (GTK_BOX (status), l, FALSE, FALSE, 4);
@@ -716,56 +667,211 @@ create_preview_canvas (GnomePrintJobPreview *mp)
 #define PRINT_AFFINE(s,a)
 #endif
 
+static void
+on_1x1_clicked (GtkMenuItem *i, GnomePrintJobPreview *mp)
+{
+	g_object_set (G_OBJECT (mp), "nx", 1, "ny", 1, NULL);
+	preview_zoom_fit_cmd (NULL, mp);
+}
+
+static void
+on_1x2_clicked (GtkMenuItem *i, GnomePrintJobPreview *mp)
+{
+	g_object_set (G_OBJECT (mp), "nx", 2, "ny", 1, NULL);
+	preview_zoom_fit_cmd (NULL, mp);
+}
+
+static void
+on_2x1_clicked (GtkMenuItem *i, GnomePrintJobPreview *mp)
+{
+	g_object_set (G_OBJECT (mp), "nx", 1, "ny", 2, NULL);
+	preview_zoom_fit_cmd (NULL, mp);
+}
+
+static void
+on_2x2_clicked (GtkMenuItem *i, GnomePrintJobPreview *mp)
+{
+	g_object_set (G_OBJECT (mp), "nx", 2, "ny", 2, NULL);
+	preview_zoom_fit_cmd (NULL, mp);
+}
+
+static void
+on_other_clicked (GtkMenuItem *i, GnomePrintJobPreview *mp)
+{
+	g_warning ("Not implemented!");
+}
+
+static void
+gpmp_multi_cmd (GtkWidget *w, GnomePrintJobPreview *mp)
+{
+	GtkWidget *m, *i;
+
+	m = gtk_menu_new ();
+	gtk_widget_show (m);
+	g_signal_connect (m, "selection_done", G_CALLBACK (gtk_widget_destroy),
+			  m);
+
+        i = gtk_menu_item_new_with_label ("1x1");
+        gtk_widget_show (i);
+	gtk_menu_attach (GTK_MENU (m), i, 0, 1, 0, 1);
+        g_signal_connect (i, "activate", G_CALLBACK (on_1x1_clicked), mp);
+
+        i = gtk_menu_item_new_with_label ("2x1");
+        gtk_widget_show (i);
+	gtk_menu_attach (GTK_MENU (m), i, 0, 1, 1, 2);
+        g_signal_connect (i, "activate", G_CALLBACK (on_2x1_clicked), mp);
+
+        i = gtk_menu_item_new_with_label ("1x2");
+        gtk_widget_show (i);
+	gtk_menu_attach (GTK_MENU (m), i, 1, 2, 0, 1);
+        g_signal_connect (i, "activate", G_CALLBACK (on_1x2_clicked), mp);
+
+        i = gtk_menu_item_new_with_label ("2x2");
+        gtk_widget_show (i);
+	gtk_menu_attach (GTK_MENU (m), i, 1, 2, 1, 2);
+        g_signal_connect (i, "activate", G_CALLBACK (on_2x2_clicked), mp);
+
+        i = gtk_menu_item_new_with_label (_("Other"));
+        gtk_widget_show (i);
+	gtk_menu_attach (GTK_MENU (m), i, 0, 2, 2, 3);
+        gtk_widget_set_sensitive (i, FALSE);
+        g_signal_connect (i, "activate", G_CALLBACK (on_other_clicked), mp);
+
+	gtk_menu_popup (GTK_MENU (m), NULL, NULL, NULL, mp, 0,
+			GDK_CURRENT_TIME);
+}
+
 static void
 create_toplevel (GnomePrintJobPreview *mp)
 {
 	GPMPPrivate *priv;
 	GtkWidget *tb;
 	gint width, height;
+	GdkGeometry hints;
+	GtkToolItem *i;
 
 	priv = (GPMPPrivate *) mp->priv;
 
-	width  = MIN (mp->paw + PAGE_PAD * 3, gdk_screen_width () - 40);
-	height = MIN (mp->pah + PAGE_PAD * 3, gdk_screen_height () - 40);
-
-	gtk_widget_set_size_request (GTK_WIDGET (mp), width, height);
-
-	/* This is not very beautiful, but works for now */
-
 	tb = gtk_toolbar_new ();
 	gtk_toolbar_set_style (GTK_TOOLBAR (tb), GTK_TOOLBAR_ICONS);
 	gtk_widget_show (tb);
 	gtk_box_pack_start (GTK_BOX (mp->vbox), tb, FALSE, FALSE, 0);
 
-	gtk_toolbar_insert_stock (GTK_TOOLBAR (tb), GTK_STOCK_PRINT,
-				  _("Prints the current file"), "", 
-				  GTK_SIGNAL_FUNC (preview_file_print_cmd), mp, -1);
+	i = gtk_tool_button_new_from_stock (GTK_STOCK_PRINT);
+	gtk_tool_item_set_tooltip (i, GTK_TOOLBAR (tb)->tooltips,
+				   _("Prints the current file"), "");
+	g_signal_connect (i, "clicked",
+			  G_CALLBACK (preview_file_print_cmd),mp);
+	gtk_widget_show (GTK_WIDGET (i));
+	gtk_toolbar_insert (GTK_TOOLBAR (tb), i, -1);
 
-	gtk_toolbar_insert_stock (GTK_TOOLBAR (tb), GTK_STOCK_CLOSE,
-				  _("Closes print preview window"), "", 
-				  GTK_SIGNAL_FUNC (preview_close_cmd), mp, -1);
+	i = gtk_tool_button_new_from_stock (GTK_STOCK_CLOSE);
+	gtk_tool_item_set_tooltip (i, GTK_TOOLBAR (tb)->tooltips, 
+				   _("Closes print preview window"), "");
+	g_signal_connect (i, "clicked",
+			  G_CALLBACK (preview_close_cmd), mp);
+	gtk_widget_show (GTK_WIDGET (i));
+	gtk_toolbar_insert (GTK_TOOLBAR (tb), i, -1);
 
-	gtk_toolbar_append_space (GTK_TOOLBAR (tb));
+	i = gtk_separator_tool_item_new ();
+	gtk_widget_show (GTK_WIDGET (i));
+	gtk_toolbar_insert (GTK_TOOLBAR (tb), i, -1);
 
-	mp->bpf = gtk_toolbar_insert_stock (GTK_TOOLBAR (tb), GTK_STOCK_GOTO_FIRST, _("Shows the first page"), "", 
-					    GTK_SIGNAL_FUNC (preview_first_page_cmd), mp, -1);
-	mp->bpp = gtk_toolbar_insert_stock (GTK_TOOLBAR (tb), GTK_STOCK_GO_BACK, _("Shows the previous page"), "", 
-					    GTK_SIGNAL_FUNC (preview_prev_page_cmd), mp, -1);
-	mp->bpn = gtk_toolbar_insert_stock (GTK_TOOLBAR (tb), GTK_STOCK_GO_FORWARD, _("Shows the next page"), "", 
-					    GTK_SIGNAL_FUNC (preview_next_page_cmd), mp, -1);
-	mp->bpl = gtk_toolbar_insert_stock (GTK_TOOLBAR (tb), GTK_STOCK_GOTO_LAST, _("Shows the last page"), "", 
-					    GTK_SIGNAL_FUNC (preview_last_page_cmd), mp, -1);
+	i = gtk_tool_button_new_from_stock (GTK_STOCK_GOTO_FIRST);
+	mp->bpf = GTK_WIDGET (i);
+	gtk_tool_item_set_tooltip (i, GTK_TOOLBAR (tb)->tooltips,
+				   _("Shows the first page"), "");
+	g_signal_connect (i, "clicked",
+			  G_CALLBACK (preview_first_page_cmd), mp);
+	gtk_widget_show (GTK_WIDGET (i));
+	gtk_toolbar_insert (GTK_TOOLBAR (tb), i, -1);
+	i = gtk_tool_button_new_from_stock (GTK_STOCK_GO_BACK);
+	mp->bpp = GTK_WIDGET (i);
+	gtk_tool_item_set_tooltip (i, GTK_TOOLBAR (tb)->tooltips,
+				   _("Shows the previous page"), "");
+	g_signal_connect (i, "clicked",
+			  G_CALLBACK (preview_prev_page_cmd), mp);
+	gtk_widget_show (GTK_WIDGET (i));
+	gtk_toolbar_insert (GTK_TOOLBAR (tb), i, -1);
+	i = gtk_tool_button_new_from_stock (GTK_STOCK_GO_FORWARD);
+	mp->bpn = GTK_WIDGET (i);
+	gtk_tool_item_set_tooltip (i, GTK_TOOLBAR (tb)->tooltips,
+				   _("Shows the next page"), "");
+	g_signal_connect (i, "clicked",
+			  G_CALLBACK (preview_next_page_cmd), mp);
+	gtk_widget_show (GTK_WIDGET (i));
+	gtk_toolbar_insert (GTK_TOOLBAR (tb), i, -1);
+	i = gtk_tool_button_new_from_stock (GTK_STOCK_GOTO_LAST);
+	mp->bpl = GTK_WIDGET (i);
+	gtk_tool_item_set_tooltip (i, GTK_TOOLBAR (tb)->tooltips,
+				    _("Shows the last page"), "");
+	g_signal_connect (i, "clicked",
+			  G_CALLBACK (preview_last_page_cmd), mp);
+	gtk_widget_show (GTK_WIDGET (i));
+	gtk_toolbar_insert (GTK_TOOLBAR (tb), i, -1);
 
-	gtk_toolbar_append_space (GTK_TOOLBAR (tb));
+	i = gtk_separator_tool_item_new ();
+	gtk_widget_show (GTK_WIDGET (i));
+	gtk_toolbar_insert (GTK_TOOLBAR (tb), i, -1);
 
-	mp->bz1 = gtk_toolbar_insert_stock (GTK_TOOLBAR (tb), GTK_STOCK_ZOOM_100, _("Zooms 1:1"), "", 
-					    GTK_SIGNAL_FUNC (preview_zoom_100_cmd), mp, -1);
-	mp->bzf = gtk_toolbar_insert_stock (GTK_TOOLBAR (tb), GTK_STOCK_ZOOM_FIT, _("Zooms to fit the whole page"), "", 
-					    GTK_SIGNAL_FUNC (preview_zoom_fit_cmd), mp, -1);
-	mp->bzi = gtk_toolbar_insert_stock (GTK_TOOLBAR (tb), GTK_STOCK_ZOOM_IN, _("Zooms the page in"), "", 
-					    GTK_SIGNAL_FUNC (gpmp_zoom_in_cmd), mp, -1);
-	mp->bzo = gtk_toolbar_insert_stock (GTK_TOOLBAR (tb), GTK_STOCK_ZOOM_OUT, _("Zooms the page out"), "", 
-					    GTK_SIGNAL_FUNC (gpmp_zoom_out_cmd), mp, -1);
+	i = gtk_tool_button_new_from_stock (GTK_STOCK_ZOOM_100);
+	mp->bz1 = GTK_WIDGET (i);
+	gtk_tool_item_set_tooltip (i, GTK_TOOLBAR (tb)->tooltips,
+				   _("Zooms 1:1"), "");
+	g_signal_connect (i, "clicked",
+			  G_CALLBACK (preview_zoom_100_cmd), mp);
+	gtk_widget_show (GTK_WIDGET (i));
+	gtk_toolbar_insert (GTK_TOOLBAR (tb), i, -1);
+	i = gtk_tool_button_new_from_stock (GTK_STOCK_ZOOM_FIT);
+	mp->bzf = GTK_WIDGET (i);
+	gtk_tool_item_set_tooltip (i, GTK_TOOLBAR (tb)->tooltips,
+				   _("Zooms to fit the whole page"), "");
+	g_signal_connect (i, "clicked",
+			  G_CALLBACK (preview_zoom_fit_cmd), mp);
+	gtk_widget_show (GTK_WIDGET (i));
+	gtk_toolbar_insert (GTK_TOOLBAR (tb), i, -1);
+	i = gtk_tool_button_new_from_stock (GTK_STOCK_ZOOM_IN);
+	mp->bzi = GTK_WIDGET (i);
+	gtk_tool_item_set_tooltip (i, GTK_TOOLBAR (tb)->tooltips,
+				   _("Zooms the page in"), "");
+	g_signal_connect (i, "clicked",
+			  G_CALLBACK (gpmp_zoom_in_cmd), mp);
+	gtk_widget_show (GTK_WIDGET (i));
+	gtk_toolbar_insert (GTK_TOOLBAR (tb), i, -1);
+	i = gtk_tool_button_new_from_stock (GTK_STOCK_ZOOM_OUT);
+	mp->bzo = GTK_WIDGET (i);
+	gtk_tool_item_set_tooltip (i, GTK_TOOLBAR (tb)->tooltips,
+				   _("Zooms the page out"), "");
+	g_signal_connect (i, "clicked",
+			  G_CALLBACK (gpmp_zoom_out_cmd), mp);
+	gtk_widget_show (GTK_WIDGET (i));
+	gtk_toolbar_insert (GTK_TOOLBAR (tb), i, -1);
+
+	i = gtk_separator_tool_item_new ();
+	gtk_widget_show (GTK_WIDGET (i));
+	gtk_toolbar_insert (GTK_TOOLBAR (tb), i, -1);
+
+	i = gtk_tool_button_new_from_stock (GTK_STOCK_MISSING_IMAGE);
+	gtk_tool_item_set_tooltip (i, GTK_TOOLBAR (tb)->tooltips,
+				   _("View multiple pages"), "");
+	g_signal_connect (i, "clicked", G_CALLBACK (gpmp_multi_cmd), mp);
+	gtk_widget_show (GTK_WIDGET (i));
+	gtk_toolbar_insert (GTK_TOOLBAR (tb), i, -1);
+
+	width  = MIN (mp->paw + PAGE_PAD * 3, gdk_screen_width () * 3 / 4);
+	height = MIN (mp->pah + PAGE_PAD * 3, gdk_screen_height () * 3 / 4);
+
+	hints.base_width = mp->paw;
+        hints.base_height = mp->pah;
+        hints.min_width = 150;
+        hints.min_height = 150;
+	
+	gtk_window_set_default_size (&(mp->window),
+				     width, 
+				     height);
+	gtk_window_set_geometry_hints 
+		(&(mp->window), priv->scrolled_window,
+		 &hints, GDK_HINT_MIN_SIZE | GDK_HINT_BASE_SIZE);
 }
 
 /*
@@ -801,11 +907,11 @@ static void
 gnome_print_job_preview_destroy (GtkObject *object)
 {
 	GnomePrintJobPreview *pmp = GNOME_PRINT_JOB_PREVIEW (object);
-	GPMPPrivate *priv = pmp->priv;
+	GPMPPrivate *priv = (GPMPPrivate *) pmp->priv;
 
-	if (priv->preview != NULL) {
-		g_object_unref (G_OBJECT (priv->preview));
-		priv->preview = NULL;
+	if (pmp->page_array != NULL) {
+		g_ptr_array_free (pmp->page_array, TRUE);
+		pmp->page_array = NULL;
 	}
 
 	if (priv->job != NULL) {
@@ -813,6 +919,11 @@ gnome_print_job_preview_destroy (GtkObject *object)
 		priv->job = NULL;
 	}
 
+	if (pmp->update_func_id) {
+		g_source_remove (pmp->update_func_id);
+		pmp->update_func_id = 0;
+	}
+
 	GTK_OBJECT_CLASS (parent_class)->destroy (object);
 }
 
@@ -833,17 +944,215 @@ gnome_print_job_preview_finalize (GObject *object)
 	G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
+static gboolean
+update_func (gpointer data)
+{
+	GnomePrintJobPreview *pmp = GNOME_PRINT_JOB_PREVIEW (data);
+
+	gnome_print_job_preview_update (pmp);
+	pmp->update_func_id = 0;
+	
+	return FALSE;
+}
+
+static void
+gnome_print_job_preview_set_nx_and_ny (GnomePrintJobPreview *pmp,
+				       gulong nx, gulong ny)
+{
+	GPtrArray *a;
+	GnomeCanvasItem *group, *item;
+	GPMPPrivate *priv;
+	guint i, col, row;
+	GObject *preview;
+
+	g_return_if_fail (GNOME_IS_PRINT_JOB_PREVIEW (pmp));
+	g_return_if_fail (nx > 0);
+	g_return_if_fail (ny > 0);
+
+	pmp->nx = nx;
+	pmp->ny = ny; 
+
+	/* Remove unnecessary pages */
+	priv = (GPMPPrivate *) pmp->priv;
+	a = pmp->page_array;
+	while (pmp->page_array->len > MIN (nx * ny, priv->pagecount)) {
+
+		/* Remove preview (if any) */
+		preview = g_object_get_data (a->pdata[a->len - 1], "preview");
+		if (preview) g_object_unref (preview);
+		g_object_set_data (a->pdata[a->len - 1], "preview", NULL);
+
+		/* Remove page */
+		gtk_object_destroy (GTK_OBJECT (a->pdata[a->len - 1]));
+		g_ptr_array_remove_index (a, a->len - 1);
+	}
+
+	/* Create pages if needed */
+	while (pmp->page_array->len < MIN (nx * ny, priv->pagecount)) {
+		group = gnome_canvas_item_new (
+			gnome_canvas_root (priv->canvas),
+			GNOME_TYPE_CANVAS_GROUP, "x", 0., "y", 0., NULL);
+		g_ptr_array_add (a, group);
+		item = gnome_canvas_item_new (GNOME_CANVAS_GROUP (group),
+			GNOME_TYPE_CANVAS_RECT, "x1", 0.0, "y1", 0.0,
+			"x2", (gdouble) pmp->paw, "y2", (gdouble) pmp->pah,
+			"fill_color", "white", "outline_color", "black",
+			"width_pixels", 1, NULL);
+		gnome_canvas_item_lower_to_bottom (item);
+		g_object_set_data (G_OBJECT (group), "page", item);
+		item = gnome_canvas_item_new (GNOME_CANVAS_GROUP (group),
+			GNOME_TYPE_CANVAS_RECT, "x1", 3.0, "y1", 3.0,
+			"x2", (gdouble) pmp->paw + 3,
+			"y2", (gdouble) pmp->pah + 3,
+			"fill_color", "black", NULL);
+		gnome_canvas_item_lower_to_bottom (item);
+	}
+
+	/* Position the pages */
+	for (i = 0; i < a->len; i++) {
+		col = i % pmp->nx;
+		row = i / pmp->nx;
+		g_object_set (a->pdata[i],
+			"x", (gdouble) col * (pmp->paw + PAGE_PAD * 2),
+			"y", (gdouble) row * (pmp->pah + PAGE_PAD * 2), NULL);
+	}
+
+	gpmp_zoom (pmp, 1., FALSE);
+
+	if (!pmp->update_func_id)
+		pmp->update_func_id = g_idle_add_full (
+			G_PRIORITY_DEFAULT_IDLE, update_func, pmp, NULL);
+}
+
+static void
+gnome_print_job_preview_update (GnomePrintJobPreview *pmp)
+{
+	GnomePrintConfig *config;
+	guint col, row, i, page;
+	GPtrArray *a;
+	gdouble transform[6];
+	ArtDRect region;
+	GnomePrintContext *preview;
+	GPMPPrivate *priv;
+	GnomeCanvasItem *group;
+
+	g_return_if_fail (GNOME_IS_PRINT_JOB_PREVIEW (pmp));
+
+	a = pmp->page_array;
+	priv = pmp->priv;
+
+	/* Show something on the pages */
+	config = gnome_print_job_get_config (priv->job);
+	for (i = 0; i < a->len; i++) {
+		col = i % pmp->nx;
+		row = i / pmp->nx;
+		page = (guint) (priv->current_page / (pmp->nx * pmp->ny))
+				* (pmp->nx * pmp->ny) + i;
+		group = a->pdata[i];
+
+		/* Remove existing preview (if any) */
+		preview = g_object_get_data (G_OBJECT (group), "preview");
+		if (preview) g_object_unref (preview);
+		g_object_set_data (G_OBJECT (group), "preview", NULL);
+
+		/* Do we need to show this page? */
+		if (page < priv->pagecount)
+			gnome_canvas_item_show (group);
+		else {
+			gnome_canvas_item_hide (group);
+			continue;
+		}
+
+		transform[0] =  1.0; transform[1] =  0.0;
+		transform[2] =  0.0; transform[3] = -1.0;
+		transform[4] =  col * (pmp->paw + PAGE_PAD * 2);
+		transform[5] =  pmp->pah + row * (pmp->pah + PAGE_PAD * 2);
+		art_affine_multiply (transform, pmp->pa2ly, transform);
+		region.x0 = region.y0 = 0.0 - PAGE_PAD;
+		region.x1 = pmp->paw + PAGE_PAD;
+		region.y1 = pmp->pah + PAGE_PAD;
+		preview = gnome_print_preview_new_full (config,
+					priv->canvas, transform, &region);
+		g_object_set_data (G_OBJECT (group), "preview", preview);
+		gnome_print_preview_theme_compliance (
+			GNOME_PRINT_PREVIEW (preview), priv->theme_compliance);
+		gnome_print_job_render_page (priv->job, preview, page, TRUE);
+	}
+	gnome_print_config_unref (config);
+
+	gnome_canvas_set_scroll_region (priv->canvas,
+		0 - PAGE_PAD, 0 - PAGE_PAD,
+		(pmp->paw + PAGE_PAD * 2) * pmp->nx + PAGE_PAD,
+		(pmp->pah + PAGE_PAD * 2) * pmp->ny + PAGE_PAD);
+}
+
+enum {
+	PROP_0,
+	PROP_NX,
+	PROP_NY
+};
+
+static void
+gnome_print_job_preview_get_property (GObject *object, guint n, GValue *v, GParamSpec *pspec)
+{
+	GnomePrintJobPreview *pmp = GNOME_PRINT_JOB_PREVIEW (object);
+
+	switch (n) {
+	case PROP_NX:
+		g_value_set_ulong (v, pmp->nx);
+		break;
+	case PROP_NY:
+		g_value_set_ulong (v, pmp->ny);
+		break;
+	default:
+		break;;
+	}
+}
+
+static void
+gnome_print_job_preview_set_property (GObject *object, guint n,
+				      const GValue *v, GParamSpec *pspec)
+{
+	GnomePrintJobPreview *pmp = GNOME_PRINT_JOB_PREVIEW (object);
+
+	switch (n) {
+	case PROP_NX:
+		gnome_print_job_preview_set_nx_and_ny (pmp,
+				g_value_get_ulong (v), pmp->ny);
+		break;
+	case PROP_NY:
+		gnome_print_job_preview_set_nx_and_ny (pmp,
+				pmp->nx, g_value_get_ulong (v));
+		break;
+	default:
+		break;
+	}
+}
+
 static void
 gnome_print_job_preview_class_init (GnomePrintJobPreviewClass *klass)
 {
 	GtkObjectClass *object_class;
+	GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
 
 	object_class = (GtkObjectClass *) klass;
 
 	parent_class = gtk_type_class (GTK_TYPE_WINDOW);
 
 	object_class->destroy = gnome_print_job_preview_destroy;
-	G_OBJECT_CLASS (object_class)->finalize = gnome_print_job_preview_finalize;
+
+	gobject_class->finalize     = gnome_print_job_preview_finalize;
+	gobject_class->set_property = gnome_print_job_preview_set_property;
+	gobject_class->get_property = gnome_print_job_preview_get_property;
+
+	g_object_class_install_property (gobject_class, PROP_NX,
+		g_param_spec_ulong ("nx", _("Number of pages horizontally"),
+			_("Number of pages horizontally"), 1, 0xffff, 1,
+			G_PARAM_READWRITE));
+	g_object_class_install_property (gobject_class, PROP_NY,
+		g_param_spec_ulong ("ny", _("Number of pages vertically"),
+			_("Number of pages vertically"), 1, 0xffff, 1,
+			G_PARAM_READWRITE));
 }
 
 static void
@@ -866,6 +1175,9 @@ gnome_print_job_preview_init (GnomePrintJobPreview *mp)
 	gtk_container_add (GTK_CONTAINER (mp), mp->vbox);
 
 	mp->zoom = 1.0;
+
+	mp->nx = mp->ny = 1;
+	mp->page_array = g_ptr_array_new ();
 }
 
 GtkWidget *
@@ -894,12 +1206,13 @@ gnome_print_job_preview_new (GnomePrintJob *gpm, const guchar *title)
 	gpmp_parse_layout (gpmp);
 	create_toplevel (gpmp);
 	create_preview_canvas (gpmp);
+	gnome_print_job_preview_set_nx_and_ny (gpmp, 1, 1);
 
 	/* this zooms to fit, once we know how big the window actually is */
 	g_signal_connect (G_OBJECT (priv->canvas), "realize",
 			  (GCallback) preview_zoom_fit_cmd, gpmp);
 
-	priv->pagecount = gpmp->pages = gnome_print_job_get_pages (gpm);
+	priv->pagecount = gnome_print_job_get_pages (gpm);
 	goto_page (gpmp, 0);
 	
 	if (priv->pagecount == 0) {
@@ -912,6 +1225,7 @@ gnome_print_job_preview_new (GnomePrintJob *gpm, const guchar *title)
 	gtk_label_set_markup_with_mnemonic (GTK_LABEL (priv->last), text);
 	g_free (text);
 
+	on_1x1_clicked (NULL, gpmp);
 	return (GtkWidget *) gpmp;
 }
 
