@@ -335,7 +335,7 @@ gpa_transport_selector_rebuild_menu (GPATransportSelector *ts)
 {
 	GtkWidget *menu, *item;
 	GPANode *option, *child;
-	guchar *def;
+	guchar *def = NULL;
 	gint pos = 0;
 	gint sel = -1;
 
@@ -390,7 +390,8 @@ gpa_transport_selector_rebuild_menu (GPATransportSelector *ts)
 	gtk_widget_show (menu);
 	gtk_option_menu_set_menu (GTK_OPTION_MENU (ts->menu), menu);
 	gtk_option_menu_set_history (GTK_OPTION_MENU (ts->menu), sel);
-	g_free (def);
+	if (def)
+		g_free (def);
 
 	gpa_transport_selector_update_widgets (ts);
 }
